@@ -242,7 +242,7 @@ pub trait ProgramVisitor {
     fn visit_conditional(
         &mut self,
         reg: &Span<Reg>,
-        val: &Span<usize>,
+        val: &Span<u64>,
         then: &Span<Stmt>,
     ) -> Result<(), Self::Error> {
         self.walk_conditional(reg, val, then)
@@ -251,7 +251,7 @@ pub trait ProgramVisitor {
     fn walk_conditional(
         &mut self,
         reg: &Span<Reg>,
-        val: &Span<usize>,
+        val: &Span<u64>,
         then: &Span<Stmt>,
     ) -> Result<(), Self::Error> {
         self.visit_stmt(then)
@@ -274,14 +274,14 @@ pub trait ProgramVisitor {
 /// struct MaxFinder;
 ///
 /// impl ExprVisitor for MaxFinder {
-///     type Output = usize;
+///     type Output = u64;
 ///
-///     fn binop(&mut self, _: Binop, a: usize, b: usize) -> usize { a.max(b) }
-///     fn unop(&mut self, _: Unop, a: usize) -> usize { a }
-///     fn lookup(&mut self, _: &Symbol) -> usize { 0 }
-///     fn pi(&mut self) -> usize { 0 }
-///     fn int(&mut self, val: usize) -> usize { val }
-///     fn real(&mut self, val: f32) -> usize { 0 }
+///     fn binop(&mut self, _: Binop, a: u64, b: u64) -> u64 { a.max(b) }
+///     fn unop(&mut self, _: Unop, a: u64) -> u64 { a }
+///     fn lookup(&mut self, _: &Symbol) -> u64 { 0 }
+///     fn pi(&mut self) -> u64 { 0 }
+///     fn int(&mut self, val: u64) -> u64 { val }
+///     fn real(&mut self, val: f32) -> u64 { 0 }
 /// }
 ///
 /// fn main() {
@@ -296,7 +296,7 @@ pub trait ExprVisitor {
     fn unop(&mut self, op: Unop, a: Self::Output) -> Self::Output;
     fn lookup(&mut self, var: &Symbol) -> Self::Output;
     fn pi(&mut self) -> Self::Output;
-    fn int(&mut self, val: usize) -> Self::Output;
+    fn int(&mut self, val: u64) -> Self::Output;
     fn real(&mut self, val: f32) -> Self::Output;
 
     fn visit_expr(&mut self, expr: &Span<Expr>) -> Self::Output {
@@ -360,7 +360,7 @@ impl<'a> ExprVisitor for FrameEvaluator<'a> {
         Ok(Value::PI)
     }
 
-    fn int(&mut self, val: usize) -> Self::Output {
+    fn int(&mut self, val: u64) -> Self::Output {
         Ok(Value::int(val as i64))
     }
 
@@ -401,8 +401,8 @@ struct Frame {
     name: Symbol,
     def_name: Symbol,
     call: Option<FileSpan>,
-    qregs: HashMap<Symbol, (usize, usize)>,
-    cregs: HashMap<Symbol, (usize, usize)>,
+    qregs: HashMap<Symbol, (u64, u64)>,
+    cregs: HashMap<Symbol, (u64, u64)>,
     params: HashMap<Symbol, Value>,
 }
 
@@ -420,7 +420,9 @@ struct Definition {
 ///
 /// This can be used by constructing one with `Linearize::new`
 /// with your chosen `GateWriter`, and then using the `ProgramVisitor`
-/// impl to call `.visit_program` on your program.
+/// impl to call `.visit_program` on your program. The `depth` argument
+/// to `Linearize::new` determines how many layers of definitions
+/// it will expand.
 ///
 /// Note that it is assumed you have type-checked your program first.
 /// If you haven't you may get garbage output / random panics. If
@@ -431,49 +433,63 @@ struct Definition {
 /// struct GatePrinter;
 ///
 /// impl GateWriter for GatePrinter {
-///     fn initialize(&mut self, _: usize, _: usize) {}
+///     type Error = std::convert::Infallible;
+/// 
+///     fn initialize(&mut self, _: &[Symbol], _: &[Symbol]) -> Result<(), Self::Error> {
+///         Ok(())
+///     }
 ///
-///     fn write_cx(&mut self, copy: usize, xor: usize) {
+///     fn write_cx(&mut self, copy: usize, xor: usize) -> Result<(), Self::Error> {
 ///         println!("cx {copy} {xor}");
+///         Ok(())
 ///     }
 ///
-///     fn write_u(&mut self, theta: Value, phi: Value, lambda: Value, reg: usize) {
+///     fn write_u(&mut self, theta: Value, phi: Value, lambda: Value, reg: usize) -> Result<(), Self::Error> {
 ///         println!("u({theta}, {phi}, {lambda}) {reg}");
+///         Ok(())
 ///     }
 ///
-///     fn write_opaque(&mut self, name: &Symbol, _: &[Value], _: &[usize]) {
-///         println!("opaque gate {}", name)
+///     fn write_opaque(&mut self, name: &Symbol, _: &[Value], _: &[usize]) -> Result<(), Self::Error> {
+///         println!("opaque gate {}", name);
+///         Ok(())
 ///     }
 ///
-///     fn write_barrier(&mut self, _: &[usize]) {}
+///     fn write_barrier(&mut self, _: &[usize]) -> Result<(), Self::Error> {
+///         Ok(())
+///     }
 ///
-///     fn write_measure(&mut self, from: usize, to: usize) {
+///     fn write_measure(&mut self, from: usize, to: usize) -> Result<(), Self::Error> {
 ///         println!("measure {} -> {}", from, to);
+///         Ok(())
 ///     }
 ///
-///     fn write_reset(&mut self, reg: usize) {
+///     fn write_reset(&mut self, reg: usize) -> Result<(), Self::Error> {
 ///         println!("reset {reg}");
+///         Ok(())
 ///     }
 ///
-///     fn start_conditional(&mut self, reg: usize, count: usize, value: usize) {
+///     fn start_conditional(&mut self, reg: usize, count: usize, value: usize) -> Result<(), Self::Error> {
 ///         println!("if ({reg}:{count} == {value}) {{");
+///         Ok(())
 ///     }
 ///
-///     fn end_conditional(&mut self) {
+///     fn end_conditional(&mut self) -> Result<(), Self::Error> {
 ///         println!("}}");
+///         Ok(())
 ///     }
 /// }
 ///
 /// fn main() {
 ///     let program = ...; // acquire a program from somewhere.
 ///     program.type_check().unwrap(); // make sure to type check.
-///     let mut l = Linearize::new(GatePrinter);
+///     let mut l = Linearize::new(GatePrinter, usize::MAX);
 ///     l.visit_program(&program).unwrap();
 /// }
 /// ```
 pub struct Linearize<T> {
-    next_qid: usize,
-    next_cid: usize,
+    next_qid: u64,
+    next_cid: u64,
+    depth: usize,
     defs: HashMap<Symbol, Rc<Definition>>,
     stack: Vec<Frame>,
     frame: Frame,
@@ -495,22 +511,25 @@ pub struct Linearize<T> {
 /// is called before and after any statements that are intended
 /// to be conditional on a classical value.
 pub trait GateWriter: Sized {
-    fn initialize(&mut self, qubits: &[Symbol], bits: &[Symbol]);
-    fn write_cx(&mut self, copy: usize, xor: usize);
-    fn write_u(&mut self, theta: Value, phi: Value, lambda: Value, reg: usize);
-    fn write_opaque(&mut self, name: &Symbol, params: &[Value], args: &[usize]);
-    fn write_barrier(&mut self, regs: &[usize]);
-    fn write_measure(&mut self, from: usize, to: usize);
-    fn write_reset(&mut self, reg: usize);
-    fn start_conditional(&mut self, reg: usize, count: usize, val: usize);
-    fn end_conditional(&mut self);
+    type Error: std::error::Error + 'static;
+
+    fn initialize(&mut self, qubits: &[Symbol], bits: &[Symbol]) -> Result<(), Self::Error>;
+    fn write_cx(&mut self, copy: usize, xor: usize) -> Result<(), Self::Error>;
+    fn write_u(&mut self, theta: Value, phi: Value, lambda: Value, reg: usize) -> Result<(), Self::Error>;
+    fn write_opaque(&mut self, name: &Symbol, params: &[Value], args: &[usize]) -> Result<(), Self::Error>;
+    fn write_barrier(&mut self, regs: &[usize]) -> Result<(), Self::Error>;
+    fn write_measure(&mut self, from: usize, to: usize) -> Result<(), Self::Error>;
+    fn write_reset(&mut self, reg: usize) -> Result<(), Self::Error>;
+    fn start_conditional(&mut self, reg: usize, count: usize, val: u64) -> Result<(), Self::Error>;
+    fn end_conditional(&mut self) -> Result<(), Self::Error>;
 }
 
-impl<T> Linearize<T> {
-    pub fn new(writer: T) -> Linearize<T> {
+impl<T: GateWriter> Linearize<T> {
+    pub fn new(writer: T, depth: usize) -> Linearize<T> {
         Linearize {
             next_qid: 0,
             next_cid: 0,
+            depth,
             defs: HashMap::new(),
             stack: Vec::new(),
             frame: Frame {
@@ -566,7 +585,7 @@ impl<T> Linearize<T> {
     // The qubit number corresponds to the base of this reference,
     // and the remaining qubits in the register are the next
     // consecutive size bits.
-    fn resolve_qreg(&self, reg: &Reg) -> (usize, usize) {
+    fn resolve_qreg(&self, reg: &Reg) -> (u64, u64) {
         match reg.index {
             None => self.frame.qregs[&reg.name],
             Some(idx) => {
@@ -576,7 +595,7 @@ impl<T> Linearize<T> {
         }
     }
 
-    fn resolve_creg(&self, reg: &Reg) -> (usize, usize) {
+    fn resolve_creg(&self, reg: &Reg) -> (u64, u64) {
         match reg.index {
             None => self.frame.cregs[&reg.name],
             Some(idx) => {
@@ -655,24 +674,30 @@ impl<T: GateWriter> ProgramVisitor for Linearize<T> {
         // Since this is first called after all register declarations,
         // now is a good time to initialize the backend.
         if !self.initialized {
-            let mut qubits = vec![Symbol::new(""); self.next_qid];
-            let mut bits = vec![Symbol::new(""); self.next_cid];
+            let mut qubits = vec![Symbol::new(""); self.next_qid as usize];
+            let mut bits = vec![Symbol::new(""); self.next_cid as usize];
 
             for (name, (base, size)) in &self.frame.qregs {
                 for offset in 0..*size {
                     let n = Symbol::new(format!("{name}[{offset}]"));
-                    qubits[*base + offset] = n;
+                    qubits[(*base + offset) as usize] = n;
                 }
             }
 
             for (name, (base, size)) in &self.frame.cregs {
                 for offset in 0..*size {
                     let n = Symbol::new(format!("{name}[{offset}]"));
-                    bits[*base + offset] = n;
+                    bits[(*base + offset) as usize] = n;
                 }
             }
 
-            self.writer.initialize(&qubits, &bits);
+            let err = self.writer.initialize(&qubits, &bits).map_err(|error| {
+                LinearizeErrorKind::WriterError {
+                    error: Box::new(error)
+                }
+            });
+            self.err(err)?;
+
             self.initialized = true;
         }
 
@@ -686,12 +711,17 @@ impl<T: GateWriter> ProgramVisitor for Linearize<T> {
         for reg in regs {
             let (base, size) = self.resolve_qreg(reg);
             for offset in 0..size {
-                args.insert(base + offset);
+                args.insert((base + offset) as usize);
             }
         }
 
         let args = args.drain().collect::<Vec<_>>();
-        self.writer.write_barrier(&args);
+        let err = self.writer.write_barrier(&args).map_err(|error| {
+            LinearizeErrorKind::WriterError {
+                error: Box::new(error)
+            }
+        });
+        self.err(err)?;
 
         Ok(())
     }
@@ -703,17 +733,38 @@ impl<T: GateWriter> ProgramVisitor for Linearize<T> {
         if fsize == tsize {
             // Many - many
             for offset in 0..fsize {
-                self.writer.write_measure(fbase + offset, tbase + offset);
+                let err = self.writer
+                    .write_measure((fbase + offset) as usize, (tbase + offset) as usize)
+                    .map_err(|error| {
+                        LinearizeErrorKind::WriterError {
+                            error: Box::new(error)
+                        }
+                    });
+                self.err(err)?;
             }
         } else if fsize == 1 {
             // One - many
             for offset in 0..tsize {
-                self.writer.write_measure(fbase, tbase + offset);
+                let err = self.writer
+                    .write_measure(fbase as usize, (tbase + offset) as usize)
+                    .map_err(|error| {
+                        LinearizeErrorKind::WriterError {
+                            error: Box::new(error)
+                        }
+                    });
+                self.err(err)?;
             }
         } else {
             // Many - one
             for offset in 0..fsize {
-                self.writer.write_measure(fbase + offset, tbase);
+                let err = self.writer
+                    .write_measure((fbase + offset) as usize, tbase as usize)
+                    .map_err(|error| {
+                        LinearizeErrorKind::WriterError {
+                            error: Box::new(error)
+                        }
+                    });
+                self.err(err)?;
             }
         }
 
@@ -723,7 +774,13 @@ impl<T: GateWriter> ProgramVisitor for Linearize<T> {
     fn visit_reset(&mut self, reg: &Span<Reg>) -> Result<(), Self::Error> {
         let (base, size) = self.resolve_qreg(reg);
         for offset in 0..size {
-            self.writer.write_reset(base + offset);
+            let err = self.writer.write_reset((base + offset) as usize)
+            .map_err(|error| {
+                LinearizeErrorKind::WriterError {
+                    error: Box::new(error)
+                }
+            });
+            self.err(err)?;
         }
 
         Ok(())
@@ -737,18 +794,54 @@ impl<T: GateWriter> ProgramVisitor for Linearize<T> {
         // distinct as the CX of two identical gates is not well defined.
         if csize == xsize {
             for offset in 0..csize {
-                self.assert_different(cbase + offset, xbase + offset, copy.span, xor.span)?;
-                self.writer.write_cx(cbase + offset, xbase + offset);
+                self.assert_different(
+                    (cbase + offset) as usize,
+                    (xbase + offset) as usize,
+                    copy.span,
+                    xor.span,
+                )?;
+                let err = self.writer
+                    .write_cx((cbase + offset) as usize, (xbase + offset) as usize)
+                    .map_err(|error| {
+                        LinearizeErrorKind::WriterError {
+                            error: Box::new(error)
+                        }
+                    });
+                self.err(err)?;
             }
         } else if csize == 1 {
             for offset in 0..xsize {
-                self.assert_different(cbase, xbase + offset, copy.span, xor.span)?;
-                self.writer.write_cx(cbase, xbase + offset);
+                self.assert_different(
+                    cbase as usize,
+                    (xbase + offset) as usize,
+                    copy.span,
+                    xor.span,
+                )?;
+                let err = self.writer
+                    .write_cx(cbase as usize, (xbase + offset) as usize)
+                    .map_err(|error| {
+                        LinearizeErrorKind::WriterError {
+                            error: Box::new(error)
+                        }
+                    });
+                self.err(err)?;
             }
         } else {
             for offset in 0..csize {
-                self.assert_different(cbase + offset, xbase, copy.span, xor.span)?;
-                self.writer.write_cx(cbase + offset, xbase);
+                self.assert_different(
+                    (cbase + offset) as usize,
+                    xbase as usize,
+                    copy.span,
+                    xor.span,
+                )?;
+                let err = self.writer
+                    .write_cx((cbase + offset) as usize, xbase as usize)
+                    .map_err(|error| {
+                        LinearizeErrorKind::WriterError {
+                            error: Box::new(error)
+                        }
+                    });
+                self.err(err)?;
             }
         }
 
@@ -794,7 +887,14 @@ impl<T: GateWriter> ProgramVisitor for Linearize<T> {
         // Now do a unitary for each referenced qubit.
         let (base, size) = self.resolve_qreg(reg);
         for offset in 0..size {
-            self.writer.write_u(theta, phi, lambda, base + offset);
+            let err = self.writer
+                .write_u(theta, phi, lambda, (base + offset) as usize)
+                .map_err(|error| {
+                    LinearizeErrorKind::WriterError {
+                        error: Box::new(error)
+                    }
+                });
+            self.err(err)?;
         }
 
         Ok(())
@@ -858,7 +958,7 @@ impl<T: GateWriter> ProgramVisitor for Linearize<T> {
         // Record the base addresses of all registers.
         let mut argsn = args
             .iter()
-            .map(|a| self.resolve_qreg(a).0)
+            .map(|a| self.resolve_qreg(a).0 as usize)
             .collect::<Vec<_>>();
 
         if def.gates.is_some() {
@@ -877,11 +977,11 @@ impl<T: GateWriter> ProgramVisitor for Linearize<T> {
             }
 
             match &def.gates {
-                Some(gates) => {
+                Some(gates) if self.stack.len() < self.depth => {
                     // This gate has a body, so process it. First insert
                     // all of the arguments as quantum registers of size one.
                     for (name, arg) in def.args.iter().zip(&argsn) {
-                        self.frame.qregs.insert(name.clone(), (*arg, 1));
+                        self.frame.qregs.insert(name.clone(), (*arg as u64, 1));
                     }
 
                     // Then recurse on the body.
@@ -889,9 +989,14 @@ impl<T: GateWriter> ProgramVisitor for Linearize<T> {
                         self.visit_stmt(stmt)?;
                     }
                 }
-                None => {
-                    // If this is opaque, write it out straight away.
-                    self.writer.write_opaque(name, &values, &argsn);
+                _ => {
+                    // If this is opaque (or we passed the depth limit), write it out straight away.
+                    let err = self.writer.write_opaque(name, &values, &argsn).map_err(|error| {
+                        LinearizeErrorKind::WriterError {
+                            error: Box::new(error)
+                        }
+                    });
+                    self.err(err)?;
                 }
             }
 
@@ -914,13 +1019,24 @@ impl<T: GateWriter> ProgramVisitor for Linearize<T> {
     fn visit_conditional(
         &mut self,
         reg: &Span<Reg>,
-        val: &Span<usize>,
+        val: &Span<u64>,
         then: &Span<Stmt>,
     ) -> Result<(), Self::Error> {
         let (base, size) = self.frame.qregs[&reg.name];
-        self.writer.start_conditional(base, size, **val);
+        let err =self.writer
+            .start_conditional(base as usize, size as usize, **val).map_err(|error| {
+                LinearizeErrorKind::WriterError {
+                    error: Box::new(error)
+                }
+            });
+        self.err(err)?;
         self.visit_stmt(then)?;
-        self.writer.end_conditional();
+        let err = self.writer.end_conditional().map_err(|error| {
+            LinearizeErrorKind::WriterError {
+                error: Box::new(error)
+            }
+        });
+        self.err(err)?;
         Ok(())
     }
 }
@@ -960,6 +1076,10 @@ pub enum LinearizeErrorKind {
         /// This argument overlaps with the other.
         b: FileSpan,
     },
+    #[error("{error}")]
+    WriterError {
+        error: Box<dyn std::error::Error>
+    }
 }
 
 #[cfg(feature = "ariadne")]
@@ -1003,6 +1123,10 @@ impl LinearizeError {
                         .with_message("This overlaps with the previous reference.")
                         .with_color(ariadne::Color::Cyan)
                         .with_order(1))
+            },
+            LinearizeErrorKind::WriterError { ref error } => {
+                Report::build(ReportKind::Error, 0usize, 0usize)
+                    .with_message(format!("{}", error))
             }
         };
 
